/* FILE: oc.cc                   -*-Mode: c++-*-
 *
 *	The OOMMF Core extension.
 *
 *	This extension provides a set of C++ classes and functions and
 * Tcl commands of general utility to all OOMMF applications.  This 
 * includes portability support for managing the differences among 
 * computing platforms, compilers, and Tcl/Tk library versions.  
 *
 * Because there are platform and compiler differences in which C++ function
 * serves as the main entry point to an application, and what arguments
 * it receives from the operating system, those entry points are provided
 * as part of this extension, so the portability burden is borne here and
 * is not passed to application writers.
 * 
 * NOTICE: Please see the file ../../LICENSE
 *
 * Last modified on: $Date: 1999/06/15 21:26:34 $
 * Last modified by: $Author: dgp $
 */

/* Header files for standard libraries */
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#if defined(__sun) && !defined(__SVR4)  // SunOS
extern "C" {
  int nice(int inc); // For SunOS
}
#endif // SunOS

/* Header file for this extension */
#include "oc.h"

/* End includes */     

/*
 * The global interpreter of the application.
 */

static Tcl_Interp *globalInterp = (Tcl_Interp *)NULL;

/*
 * Do we fork a child process to do the work? (Default = no)
 *  -- Meaningful only when (SYSTEM_TYPE == UNIX)
 */
static BOOL forkflag = 0;

/*
 * Do we load the Tk extension? (Default = no)
 */
static BOOL use_tk = 0;

/*
 * The following static strings contain Tcl scripts which are evaluated
 * to provide a small library of routines needed to accomplish the task
 * of finding the initialization Tcl scripts for the Oc extension.
 *
 * NOTE: There must be several separate strings because some compilers
 * place a rather small limit on the maximum length of a string literal.
 */

// Fix for bug in Tk_Init in Tk 4.2
static char patchScript0[] =
"if {[string match 7.6 $tcl_version] && ![info exists tk_library]} {\n\
   set tk_library [file join [file dirname $tcl_library] tk4.2]\n\
}";

// NOMAC: Assumes '..' means 'parent directory'
static char initScript0[] =
"proc Oc_DirectPathname { pathname } {\n\
    global Oc_DirectPathnameCache\n\
    set canCdTo [file dirname $pathname]\n\
    set rest [file tail $pathname]\n\
    switch -exact -- $rest {. - .. {\n\
        set canCdTo [file join $canCdTo $rest]\n\
        set $rest {}\n\
    }}\n\
    if {[string match absolute [file pathtype $canCdTo]]} {\n\
        set index $canCdTo\n\
    } else {\n\
        set index [file join [pwd] $canCdTo]\n\
    }\n\
    if {[info exists Oc_DirectPathnameCache($index)]} {\n\
        return [file join $Oc_DirectPathnameCache($index) $rest]\n\
    }\n\
    while {![file executable $canCdTo]} {\n\
        switch -exact -- [file tail $canCdTo] {\n\
            {} {set Oc_DirectPathnameCache($index) $canCdTo\n\
                return [file join $Oc_DirectPathnameCache($index) $rest]}\n\
            . {} .. {\n\
                set Oc_DirectPathnameCache($index) [file dirname \\\n\
                        [Oc_DirectPathname [file dirname $canCdTo]]]\n\
                return [file join $Oc_DirectPathnameCache($index) $rest]}\n\
            default {set rest [file join [file tail $canCdTo] $rest]}\n\
        }\n\
        set canCdTo [file dirname $canCdTo]\n\
    }\n\
    set savedir [pwd]\n\
    if {[catch {cd $canCdTo} msg]} {\n\
        error \"Can't determine direct pathname for \n\
$pathname:\n\
	$msg\"\n\
    }\n\
    set Oc_DirectPathnameCache($index) [pwd]\n\
    cd $savedir\n\
    return [file join $Oc_DirectPathnameCache($index) $rest]\n\
}";

static char initScript1[] =
"proc Oc_AppendUniqueReadableDirectory {listref dir} {\n\
    upvar $listref list\n\
    if {[file isdirectory $dir] && [file readable $dir]} {\n\
        set dir [Oc_DirectPathname $dir]\n\
        if {[lsearch -exact $list $dir] < 0} {\n\
            lappend list $dir\n\
        }\n\
    }\n\
}";

static char initScript2[] =
"proc Oc_NameOfExecutable {} {\n\
    global oc\n\
    catch {return $oc(nameOfExecutable)}\n\
    set oc(nameOfExecutable) [Oc_DirectPathname\\\n\
            [info nameofexecutable]]\n\
    return $oc(nameOfExecutable)\n\
}";

static char initScript3[] =
"proc Oc_ScriptPath {Pkg ver} {\n\
    global env \n\
    set dirs {}\n\
    set envVar [string toupper $Pkg]_LIBRARY\n\
    if [info exists env($envVar)] {\n\
        Oc_AppendUniqueReadableDirectory dirs $env($envVar)\n\
    }\n\
    set pkg [string tolower $Pkg]\n\
    set tclVar ${pkg}_library\n\
    if [info exists $tclVar] {\n\
        Oc_AppendUniqueReadableDirectory dirs [set $tclVar]\n\
    }\n\
    set execName [Oc_NameOfExecutable] \n\
    set parDir [file dirname [file dirname $execName]]\n\
    Oc_AppendUniqueReadableDirectory dirs \\\n\
            [file join $parDir lib ext $pkg$ver]\n\
    set ggparDir [file dirname [file dirname $parDir]]\n\
    Oc_AppendUniqueReadableDirectory dirs \\\n\
            [file join $ggparDir ext $pkg$ver]\n\
    Oc_AppendUniqueReadableDirectory dirs \\\n\
            [file join $ggparDir ext $pkg]\n\
    Oc_AppendUniqueReadableDirectory dirs $parDir\n\
    return $dirs\n\
 }";

static char initScript[] =
"proc Oc_Init {} {\n\
    rename Oc_Init {}\n\
    proc Oc_InitScript {Pkg ver} {\n\
        set pkg [string tolower $Pkg]\n\
        global ${pkg}_library\n\
        set dirs [Oc_ScriptPath $Pkg $ver]\n\
        foreach i $dirs {\n\
            set ${pkg}_library $i\n\
            set script [file join $i $pkg.tcl]\n\
            if {[file isfile $script] && [file readable $script]} {\n\
                uplevel #0 [list source $script]\n\
                return\n\
            }\n\
        }\n\
        set msg \"\nCan't find the '$Pkg' initialization script, $pkg.tcl,\"\n\
        append msg \"\nin the following directories:\n\n\"\n\
        append msg \"[join $dirs {\n}]\n\n\"\n\
        append msg \"$Pkg probably hasn't been \"\n\
        append msg \"installed properly.\n\n\"\n\
        append msg \"Either (re-)install $Pkg, \"\n\
        append msg \"or set the environment variable \"\n\
        append msg \"[string toupper $Pkg]_LIBRARY\nto the name \"\n\
        append msg \"of a directory containing $pkg.tcl.\"\n\
        return -code error $msg\n\
    }\n\
    global oc_version\n\
    Oc_InitScript Oc $oc_version\n\
}\n\
Oc_Init";

int
Oc_RegisterCommand(Tcl_Interp* interp, const char* name,
                   Tcl_CmdProc* cmd)
{
    Oc_AutoBuf ab;
    if(Tcl_CreateCommand(interp,ab(name),cmd,NULL,NULL)==NULL) {
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "Unable to register command ->",
                name, "<- with Tcl interpreter", (char *) NULL);
        return TCL_ERROR;
    }
    return TCL_OK;
}

static int
IgnoreSignal(int signo)
{
  //  1) SIGHUP       2) SIGINT       3) SIGQUIT      4) SIGILL
  //  5) SIGTRAP      6) SIGIOT       7) SIGEMT       8) SIGFPE
  //  9) SIGKILL     10) SIGBUS      11) SIGSEGV     12) SIGSYS
  // 13) SIGPIPE     14) SIGALRM     15) SIGTERM     16) SIGURG
  // 17) SIGSTOP     18) SIGTSTP     19) SIGCONT     20) SIGCHLD
  // 21) SIGTTIN     22) SIGTTOU     23) SIGIO       24) SIGXCPU
  // 25) SIGXFSZ     26) SIGVTALRM   27) SIGPROF     28) SIGWINCH
  // 29) SIGPWR      30) SIGUSR1     31) SIGUSR2
  // Under Windows, only the following subset is apparently available:
  //   SIGINT, SIGILL, SIGFPE, SIGSEGV, SIGTERM, SIGBREAK, SIGABRT
  // Unconfirmed reports claim Windows will crash if signal() is sent
  // a value outside this set.  Note that signal() will correctly
  // return a TCL_ERROR if the value is outside the known set of
  // signals.
#if 0 && (SYSTEM_TYPE==WINDOWS)
  if(signo!=SIGINT && signo!=SIGILL && signo!=SIGFPE
     && signo!=SIGSEGV && signo!=SIGTERM && signo!=SIGBREAK
     && signo!=SIGABRT)
    return TCL_ERROR;
#endif
#ifndef __GNUC__
  if( signal(signo,SIG_IGN) == SIG_ERR ) return TCL_ERROR;
#else
  if( (void*)signal(signo,SIG_IGN) == (void*)SIG_ERR ) return TCL_ERROR;
  /// The (void*) casts above are to work around a gcc-2.7.2.2 (and
  /// others?) bug that shows up on platforms with a non-ANSI signal(),
  /// e.g., SunOS.  In that circumstance, the return value from signal()
  /// is treated as void (*)(...) instead of void (*)(), i.e.,
  /// void (*)(void).
#endif
  return TCL_OK;
}

static int
Oc_IgnoreSignal(ClientData, Tcl_Interp *interp, int argc, char **argv) 
{
    Tcl_ResetResult(interp);
    if (argc != 2) {
        Tcl_AppendResult(interp, argv[0], " must be called with"
            " 1 argument: signal", (char *) NULL);
        return TCL_ERROR;
    }
    return IgnoreSignal(atoi(argv[1]));
}

static void MakeNice()
{
#if (SYSTEM_TYPE==WINDOWS)
    // Set solver to run at idle priority
    SetPriorityClass(OpenProcess(PROCESS_SET_INFORMATION,TRUE,
                                 GetCurrentProcessId()),
                     NICE_DEFAULT);
#else
    nice(NICE_DEFAULT);
#endif
}

static int
Oc_MakeNice(ClientData, Tcl_Interp *interp, int argc, char **argv) 
{
    Tcl_ResetResult(interp);
    if (argc != 1) {
        Tcl_AppendResult(interp, argv[0], " takes no arguments", (char *) NULL);
        return TCL_ERROR;
    }
    MakeNice();
    return TCL_OK;
}

Tcl_Interp *
Oc_GlobalInterpreter()
{
    return globalInterp;
}

static void
ClearGlobalInterpreter(ClientData, Tcl_Interp *interp)
{
    if (interp == globalInterp) {
        globalInterp = (Tcl_Interp *) NULL;
    } else {
      Oc_AutoBuf ab;
        panic(ab("Global interpreter mismatch!"));
    }
}

void
Oc_SetUseTkFlag(int val)
{
    use_tk = val;
}

int 
Oc_Init(Tcl_Interp *interp)
{
    Oc_AutoBuf ab, ab1, ab2;

    if (globalInterp == (Tcl_Interp *)NULL) {
        globalInterp = interp;
        Tcl_CallWhenDeleted(interp, 
		(Tcl_InterpDeleteProc *) ClearGlobalInterpreter,
		(ClientData) NULL);
    }

    // Normally the function Oc_Init(Tcl_Interp *interp) will be called by 
    // the application initialization function of the application using this 
    // extension.  The conventional name for that function is 
    // Tcl_AppInit(Tcl_Interp *interp).  The argument 'interp' passed to 
    // Tcl_AppInit (and passed on to Oc_Init) was presumably returned
    // by a call to Tcl_CreateInterp(), usually within one of the functions
    // Tcl_Main() or Tk_Main().  The nice thing about this, is that
    // Tcl_CreateInterp() contains a call to TclPlatformInit(), an internal
    // routine inside the Tcl library which takes care of basic 
    // platform-specific initialization.  This means that although Tcl_Init
    // hasn't yet been called (and therefore the interpreter is not fully
    // initialized), the interpreter is usable in a platform-independent
    // manner, and the Tcl variables tcl_pkgPath, tcl_library, tcl_platform,
    // tcl_patchLevel, tcl_version, tcl_precision and env(HOME) have already 
    // been set.  Also the "Tcl" package has already been provided.
    //
    // Until the Oc_Log cross-platform facility for displaying messages to 
    // the user is supported by evaluating the Tcl portion of this extension 
    // below, we must fall back on calls to panic().


    // Initialize Tcl.  
#ifdef CONFIG_TCL_LIBRARY
    // Set environment variable TCL_LIBRARY, unless already set.
    // We set this instead of tcl_library, because the tcl_library
    // is not inherited into slave interpreters.
    if(Tcl_GetVar2(interp,ab1("env"),ab2("TCL_LIBRARY"),
		   TCL_GLOBAL_ONLY)==NULL) {
      Tcl_SetVar2(interp,ab1("env"),ab2("TCL_LIBRARY"),
		  OC_STRINGIFY(CONFIG_TCL_LIBRARY),
		  TCL_GLOBAL_ONLY);
    }
#endif
    if (Tcl_Init(interp) != TCL_OK) {
        if(strncmp("Can't find a usable",Tcl_GetStringResult(interp),19)==0) {
            panic(ab("Tcl initialization error:\n\n%s\n\n"
                    "As a workaround, set the environment variable\n"
                    "TCL_LIBRARY to the name of a directory\n"
                    "containing an error-free init.tcl."),
                    Tcl_GetStringResult(interp));
        } else {
            panic(ab("Tcl initialization error:\n\n%s"),
                    Tcl_GetStringResult(interp));
        }
    }

    if (Nullchannel_Init(interp) != TCL_OK) {
	panic(ab("Nullchannel initialization error:\n\n%s"),
		Tcl_GetStringResult(interp));
    }

    // Patch a bug in Tk 4.2
    if (Tcl_Eval(interp, ab(patchScript0)) != TCL_OK) {
       panic(ab("Tcl/Tk initialization error:\n\n"
               "Unable to patch installation bug in Tk 4.2\n\n"
               "Patch script returned error:%s"), Tcl_GetStringResult(interp));
    }

    // Initialize Tk, if requested.
    if (use_tk) {
        if (Tk_Init(interp) != TCL_OK) {
            if(strncmp("Can't find a usable",
                    Tcl_GetStringResult(interp),19)==0) {
                panic(ab("Tk initialization error:\n\n%s\n\n"
                        "As a workaround, set the environment variable\n"
                        "TK_LIBRARY to the name of a directory\n"
                        "containing an error-free tk.tcl."),
                        Tcl_GetStringResult(interp));
            } else {
                panic(ab("Tk initialization error:\n\n%s"),
                        Tcl_GetStringResult(interp));
            }
        }
 
        // Set up so that [info loaded] result includes Tk
        // Tk_SafeInit() became available in Tk version 8
#if (TK_MAJOR_VERSION >= 8)
        Tcl_StaticPackage(interp,ab("Tk"), Tk_Init, Tk_SafeInit);
#else // TK_MAJOR_VERSION < 8
        Tcl_StaticPackage(interp,ab("Tk"), Tk_Init, 
                (Tcl_PackageInitProc *) NULL);
#endif
    }

    // Register C++-based commands of this extension
#if (SYSTEM_TYPE==WINDOWS)
    Oc_RegisterCommand(interp,"Oc_WindowsMessageBox",
                       Oc_WindowsMessageBoxCmd);
#endif
    Oc_RegisterCommand(interp, "Oc_IgnoreSignal",
	    (Tcl_CmdProc *) Oc_IgnoreSignal);
    Oc_RegisterCommand(interp, "Oc_MakeNice",
	    (Tcl_CmdProc *) Oc_MakeNice);
    Oc_RegisterCommand(interp,"Oc_SetPanicHeader",Oc_SetPanicHeaderCmd);

    // Evaluate the Tcl portion of this extension
    Tcl_SetVar(interp,ab1("oc_version"),ab2(OC_VERSION),TCL_GLOBAL_ONLY);
    if (Tcl_Eval(interp, ab(initScript0)) != TCL_OK
            || Tcl_Eval(interp, ab(initScript1)) != TCL_OK
            || Tcl_Eval(interp, ab(initScript2)) != TCL_OK
            || Tcl_Eval(interp, ab(initScript3)) != TCL_OK) {
        panic(ab("Error in extension 'Oc':\n\n"
                "Error evaluating pre-initialization scripts:\n%s"),
                Tcl_GetStringResult(interp));
    }
    if (Tcl_Eval(interp, ab(initScript)) != TCL_OK) {
        panic(ab("%s"),Tcl_GetVar(interp, ab1("errorInfo"), TCL_GLOBAL_ONLY));
    }

    // At this stage Oc_Log is available for displaying error messages
    // to the user.

    // Make Tcl build version available
    Tcl_SetVar(interp,ab1("omfTclBuildVersion"), ab2(TCL_PATCH_LEVEL),
            TCL_GLOBAL_ONLY);

    return TCL_OK;
}

/* 
 * Removes argument "remove_index" from argument list by
 * translating all arguments after it down one spot and
 * decrementing argc.
 */
static void 
remove_arg(int &argc, char **argv, int remove_index)
{ 
    if(remove_index<0 || remove_index>=argc) return;
    for(int j=remove_index+1;j<argc;j++) {
        argv[j-1]=argv[j];
    }
    argc--;
}

static void 
ExtractOptions(int &argc, char **argv)
{
    int i=1;
    while(i<argc) {
        if(strcmp(argv[i],"--")==0){
            // End flag processing
            break;
        }

        if(argv[i][0]!='-') {
            // Not an option flag. Under Windows, should one
            // use '/' instead of '-' ??? 
            i++;
            continue;
        }

        char *optflag=&(argv[i][1]);
        if(*optflag=='-') optflag++;  // Treat "-" and "--" the same

        // Known options
        //	Should this processing be case sensitive ???
        if (strcmp(optflag,"fork")==0) {
            forkflag=1;
            remove_arg(argc,argv,i);
        }
        else if(strcmp(optflag,"nofork")==0 || strcmp(optflag,"no_fork")==0) {
            forkflag=0;
            remove_arg(argc,argv,i);
        }
        else {
            // Anything else is left on the command line untouched, for
            // possible use by the shell script
            i++;
        }
    }
}

#if (SYSTEM_TYPE==WINDOWS)
/*
 * The main entry point for OOMMF applications on Windows platforms.
 */
int APIENTRY 
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine,
        int nCmdShow)
{
    // Windows doesn't pass the application name as the first word on the
    // command line like Unix does.  Instead a system call provides the
    // application name (as a full pathname).
    // NOTE: This fixed-length buffer should be long enough even for a full
    // pathname.  Dynamic allocation from the heap is avoided since
    // Tcl_Main and Tk_Main never return, and we'll have no chance to 
    // free the memory.  It's not clear whether Windows will free memory
    // when the process exits.
    char *p,exename[1024];
    GetModuleFileName(NULL, exename, sizeof(exename));
    // Translate '\' to '/'
    for (p = exename; *p != '\0'; p++) {
	if (*p == '\\') *p = '/';
    }
    for (p = lpszCmdLine; *p != '\0'; p++) {
	if (*p == '\\') *p = '/';
    }

    // Initialize the panic() function to use our customized
    // message display routines.
    Oc_InitPanic(exename);

    // Tcl_FindExecutable does some double-secret special
    // initializations without which some Tcl C APIs (like
    // Tcl_CreateChannel()) will crash.  So, call it here.
    Tcl_FindExecutable(exename);

    // Query the OS whether we were launched with std channels
    STARTUPINFO startInfo;
    GetStartupInfo(&startInfo);

    // Programs with entry point WinMain() which are launched
    // from the operating system do not have any standard channels.
    // When such programs are launched as subprocesses by another
    // program via the CreateProcess() system call, the parent
    // process can pass standard channels to the subprocess.  The
    // startInfo structure retrieved from the OS above contains
    // information about what instructions the parent process 
    // supplied for this program.  If the STARTF_USESTDHANDLES bit
    // of the flag array startInfo.dwFlags is set, then the parent
    // process provided standard channels to this process.  In that
    // case, it is safe to call the Tcl library function
    // Tcl_GetStdChannel().  If the STARTF_USESTDHANDLES flag is
    // not set, or if a call to Tcl_GetStdChannel() reveals that
    // there is no usable handle (by returning NULL), we must 
    // construct our own standard channels.  If we do not, then
    // any Tcl commands which would operate on the standard channels
    // will raise an error.  Here, when standard channels must be 
    // created, we create "null" channels -- analogous to /dev/null
    // on Unix systems.
    // A further improvement would be to find a way to make 
    // the standard Tcl channels and the C++ variables stdin, stdout, 
    // and stderr refer to the same devices.  
    //   Note by mjd 981010: On Windows NT/Alpha, and according to the
    // Tcl docs, it appears that Tcl_GetStdChannel as used below will
    // always a non-NULL value.  The returned channel may not be a valid
    // channel, though.  Subsequent i/o on such a channel returns the
    // error "Bad file number".  However, on this platform at least,
    // Tcl_Tell can be used to detect this condition---it returns 0
    // on good channels, -1 on bad.
    Oc_AutoBuf ab;
    Tcl_Channel channel;
    int redo_stdin=1,redo_stdout=1,redo_stderr=1;
    if (startInfo.dwFlags & STARTF_USESTDHANDLES) {
      // startInfo indicates that we have valid standard
      // channels.  Double check this.
      if ((channel=Tcl_GetStdChannel(TCL_STDIN))!=NULL) {
	if (Tcl_Tell(channel)>=0) {
	  redo_stdin = 0;  // Good channel!
	} else {
	  Tcl_UnregisterChannel(NULL,channel);
	}
      }
      if ((channel=Tcl_GetStdChannel(TCL_STDOUT))!=NULL) {
	if (Tcl_Tell(channel)>=0) {
	  redo_stdout = 0;  // Good channel!
	} else {
	  Tcl_UnregisterChannel(NULL,channel);
	}
      }
      if ((channel=Tcl_GetStdChannel(TCL_STDERR))!=NULL) {
	if (Tcl_Tell(channel)>=0) {
	  redo_stderr = 0;  // Good channel!
	} else {
	  Tcl_UnregisterChannel(NULL,channel);
	}
      }
    }
    if(redo_stdin) {
      channel = Nullchannel_Open();
      if (channel == NULL) {
	panic(ab("Couldn't create a standard input channel"));
      }
      Tcl_SetStdChannel(channel, TCL_STDIN);
    }
    if(redo_stdout) {
      channel = Nullchannel_Open();
      if (channel == NULL) {
	panic(ab("Couldn't create a standard output channel"));
      }
      Tcl_SetStdChannel(channel, TCL_STDOUT);
    }
    if(redo_stderr) {
      channel = Nullchannel_Open();
      if (channel == NULL) {
	panic(ab("Couldn't create a standard error channel"));
      }
      Tcl_SetStdChannel(channel, TCL_STDERR);
    }

    // Windows doesn't tokenize the command line
    int argc;
    char *argv[256],**argv_temp;
    char cmdline[4096];
    if(Tcl_SplitList(NULL,lpszCmdLine,&argc,&argv_temp)!=TCL_OK) {
      panic("Bad command line: %s",lpszCmdLine);
    }
    if(argc+1>sizeof(argv)/sizeof(argv[0])) {
      panic(ab("Too many command line parameters: %d>%d"),
	       argc+1,sizeof(argv)/sizeof(argv[0]));
    }

    // Copy split string out of temporary memory, and setup
    // argv pointers.
    argv[0]=exename;
    int i,total_length=0;
    for(i=0;i<argc;i++) {
      argv[i+1]=cmdline+total_length;
      total_length+=strlen(argv_temp[i])+1;
      if(total_length>sizeof(cmdline)) {
	panic(ab("Parsed command line too long (%d)"),sizeof(cmdline));
      }
      strcpy(argv[i+1],argv_temp[i]);
    }
    argc++;
    Tcl_Free((char *)argv_temp);

    // Now we can pretend a Unix system gave us argc and argv
    ExtractOptions(argc,argv);

    return Oc_Main(argc, argv);
}

#elif (SYSTEM_TYPE == UNIX)

/*
 * The main entry point for OOMMF applications on Unix platforms.
 */
int 
main(int argc, char **argv)
{
    // Tcl_FindExecutable does some double-secret special
    // initializations without which some Tcl C APIs (like
    // Tcl_CreateChannel()) will crash.  So, call it here to be safe.
    Tcl_FindExecutable(argv[0]);

    Oc_InitPanic(argv[0]);
    ExtractOptions(argc,argv);

    // Fork off a child process to do all the work.
    // Parent process terminates and user gets back shell prompt
    if(!forkflag || fork()==0) { // Fork only occurs if forkflag!=0
        // Child or non-forked process

        // In child process, disable stdin, stdout
        if (forkflag) {
            freopen("/dev/null","r",stdin);
            freopen("/dev/null","a",stdout);
        }

        return Oc_Main(argc, argv);
    }

    // Parent process exit
    return 0;
}

#else
#error "Unknown SYSTEM_TYPE"
#endif // SYSTEM_TYPE == WINDOWS
